# This file is automatically generated on build, do not edit!
# Do not commit this file to source control!

__all__ = [
    "__version__",
    "__version_tuple__",
    "version",
    "version_tuple"
]

version: str
__version__: str
__version_tuple__: tuple[int | str, ...]
version_tuple: tuple[int | str, ...]

__version__ = version = '4.0.0'
__version_tuple__ = version_tuple = (4, 0, 0)

